using System;
using System.Collections;
using System.Collections.Generic;

using Org.BouncyCastle.Asn1;
using Org.BouncyCastle.Asn1.CryptoPro;
using Org.BouncyCastle.Asn1.Eac;
using Org.BouncyCastle.Asn1.Nist;
using Org.BouncyCastle.Asn1.Oiw;
using Org.BouncyCastle.Asn1.Pkcs;
using Org.BouncyCastle.Asn1.TeleTrust;
using Org.BouncyCastle.Asn1.X509;
using Org.BouncyCastle.Asn1.X9;
using Org.BouncyCastle.Cert;
using Org.BouncyCastle.Utilities;

namespace Org.BouncyCastle.Cms
{
    internal static class CmsSignedHelper
    {
        private static readonly IDictionary m_digestAlgs = Platform.CreateHashtable();

        private static void AddDigest(string name, params DerObjectIdentifier[] oids)
        {
            foreach (var oid in oids)
            {
                m_digestAlgs.Add(oid.Id, name);
            }
        }

        private static void AddEntries(DerObjectIdentifier oid, string digest, string encryption)
		{
			string alias = oid.Id;
			m_digestAlgs.Add(alias, digest);
		}

		static CmsSignedHelper()
		{
            AddDigest("GOST3411",
                CryptoProObjectIdentifiers.GostR3411,
                new DerObjectIdentifier("1.3.6.1.4.1.5849.1.2.1"));
            AddDigest("MD2",
                PkcsObjectIdentifiers.MD2,
                PkcsObjectIdentifiers.MD2WithRsaEncryption);
            AddDigest("MD4",
                OiwObjectIdentifiers.MD4WithRsa,
                OiwObjectIdentifiers.MD4WithRsaEncryption,
                PkcsObjectIdentifiers.MD4,
                PkcsObjectIdentifiers.MD4WithRsaEncryption);
            AddDigest("MD5",
                OiwObjectIdentifiers.MD5WithRsa,
                PkcsObjectIdentifiers.MD5,
                PkcsObjectIdentifiers.MD5WithRsaEncryption);
            AddDigest("RIPEMD128",
                TeleTrusTObjectIdentifiers.RipeMD128);
            AddDigest("RIPEMD160",
                TeleTrusTObjectIdentifiers.RipeMD160);
            AddDigest("RIPEMD256",
                TeleTrusTObjectIdentifiers.RipeMD256);
            AddDigest("SHA1",
                EacObjectIdentifiers.id_TA_ECDSA_SHA_1,
                EacObjectIdentifiers.id_TA_RSA_PSS_SHA_1,
                EacObjectIdentifiers.id_TA_RSA_v1_5_SHA_1,
                OiwObjectIdentifiers.DsaWithSha1,
                OiwObjectIdentifiers.IdSha1,
                OiwObjectIdentifiers.Sha1WithRsa,
                PkcsObjectIdentifiers.Sha1WithRsaEncryption,
                X9ObjectIdentifiers.ECDsaWithSha1,
                X9ObjectIdentifiers.IdDsaWithSha1);
            AddDigest("SHA224",
                EacObjectIdentifiers.id_TA_ECDSA_SHA_224,
                NistObjectIdentifiers.DsaWithSha224,
                NistObjectIdentifiers.IdSha224,
                PkcsObjectIdentifiers.Sha224WithRsaEncryption,
                X9ObjectIdentifiers.ECDsaWithSha224);
            AddDigest("SHA256",
                EacObjectIdentifiers.id_TA_ECDSA_SHA_256,
                EacObjectIdentifiers.id_TA_RSA_PSS_SHA_256,
                EacObjectIdentifiers.id_TA_RSA_v1_5_SHA_256,
                NistObjectIdentifiers.DsaWithSha256,
                NistObjectIdentifiers.IdSha256,
                PkcsObjectIdentifiers.Sha256WithRsaEncryption,
                X9ObjectIdentifiers.ECDsaWithSha256);
            AddDigest("SHA384",
                EacObjectIdentifiers.id_TA_ECDSA_SHA_384,
                NistObjectIdentifiers.DsaWithSha384,
                NistObjectIdentifiers.IdSha384,
                PkcsObjectIdentifiers.Sha384WithRsaEncryption,
                X9ObjectIdentifiers.ECDsaWithSha384);
            AddDigest("SHA512",
                EacObjectIdentifiers.id_TA_ECDSA_SHA_512,
                NistObjectIdentifiers.DsaWithSha512,
                NistObjectIdentifiers.IdSha512,
                PkcsObjectIdentifiers.Sha512WithRsaEncryption,
                X9ObjectIdentifiers.ECDsaWithSha512);
		}

        private static void AddCertsFromSet(IList<X509Certificate> certs, Asn1Set certSet)
        {
            foreach (Asn1Encodable ae in certSet)
            {
                try
                {
                    Asn1Object obj = ae.ToAsn1Object();

                    if (obj is Asn1Sequence)
                    {
                        certs.Add(new X509Certificate(obj.GetEncoded()));
                    }
                }
                catch (Exception ex)
                {
                    throw new CmsException("can't re-encode certificate!", ex);
                }
            }
        }

        private static void AddCrlsFromSet(IList<X509Crl> crls, Asn1Set crlSet)
        {
            foreach (Asn1Encodable ae in crlSet)
            {
                try
                {
                    crls.Add(new X509Crl(ae.GetEncoded()));
                }
                catch (Exception ex)
                {
                    throw new CmsException("can't re-encode CRL!", ex);
                }
            }
        }

        internal static IStore<X509V2AttributeCertificate> CreateAttributeStore(Asn1Set certSet)
		{
			IList<X509V2AttributeCertificate> certs = new List<X509V2AttributeCertificate>();

			if (certSet != null)
			{
				foreach (Asn1Encodable ae in certSet)
				{
					try
					{
						Asn1Object obj = ae.ToAsn1Object();

						if (obj is Asn1TaggedObject)
						{
							Asn1TaggedObject tagged = (Asn1TaggedObject)obj;

							if (tagged.TagNo == 2)
							{
								certs.Add(
									new X509V2AttributeCertificate(
										Asn1Sequence.GetInstance(tagged, false).GetEncoded()));
							}
						}
					}
					catch (Exception ex)
					{
						throw new CmsException("can't re-encode attribute certificate!", ex);
					}
				}
			}

			try
			{
				return new CollectionStore<X509V2AttributeCertificate>(certs);
			}
			catch (ArgumentException e)
			{
				throw new CmsException("can't setup the X509Store", e);
			}
		}

		internal static IStore<X509Certificate> CreateCertificateStore(Asn1Set certSet)
		{
            IList<X509Certificate> certs = new List<X509Certificate>();
            if (certSet != null)
			{
				AddCertsFromSet(certs, certSet);
			}

			try
			{
                return new CollectionStore<X509Certificate>(certs);
            }
			catch (ArgumentException e)
			{
				throw new CmsException("can't setup the X509Store", e);
			}
		}

		internal static IStore<X509Crl> CreateCrlStore(Asn1Set crlSet)
		{
            IList<X509Crl> crls = new List<X509Crl>();
            if (crlSet != null)
			{
				AddCrlsFromSet(crls, crlSet);
			}

			try
			{
                return new CollectionStore<X509Crl>(crls);
            }
			catch (ArgumentException e)
			{
				throw new CmsException("can't setup the X509Store", e);
			}
		}

		internal static AlgorithmIdentifier FixDigestAlgID(AlgorithmIdentifier algID)
		{
			if (algID.Parameters == null)
                return new AlgorithmIdentifier(algID.Algorithm, DerNull.Instance);

			return algID;
		}

        internal static string GetDigestAlgName(string digestAlgOid)
        {
            string algName = (string)m_digestAlgs[digestAlgOid];
            if (algName != null)
                return algName;

            return digestAlgOid;
        }
    }
}
