﻿using System;

using Org.BouncyCastle.Asn1.Pkcs;
using Org.BouncyCastle.Asn1.X509;
using Org.BouncyCastle.Asn1.X9;
using Org.BouncyCastle.Cert;
using Org.BouncyCastle.Cms;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Asymmetric;
using Org.BouncyCastle.Crypto.Fips;

namespace Org.BouncyCastle.Operators
{
    public class PkixSignerInformationVerifierProvider
        : ISignerInformationVerifierProvider
    {
        private readonly IAsymmetricPublicKey m_publicKey;
        private readonly X509Certificate m_certificate;

        public PkixSignerInformationVerifierProvider(IAsymmetricPublicKey publicKey)
        {
            m_publicKey = publicKey;
            m_certificate = null;
        }

        public PkixSignerInformationVerifierProvider(X509Certificate cert)
        {
            m_publicKey = cert.GetPublicKey();
            m_certificate = cert;
        }

        public bool IsRawSigner => false;

        public IDigestFactory<AlgorithmIdentifier> CreateDigestFactory(AlgorithmIdentifier digestAlgorithmID) =>
            new PkixDigestFactory(digestAlgorithmID);

        public IVerifierFactory<AlgorithmIdentifier> CreateVerifierFactory(AlgorithmIdentifier signatureAlgorithmID,
            AlgorithmIdentifier digestAlgorithmID)
        {
            var sigAlgOid = signatureAlgorithmID.Algorithm;

            if (m_publicKey is AsymmetricRsaPublicKey rsaKey)
            {
                IVerifierFactoryService verifierService = CryptoServicesRegistrar.CreateService(rsaKey);

                if (PkcsObjectIdentifiers.IdRsassaPss.Equals(sigAlgOid))
                {
                    RsassaPssParameters sigParams = RsassaPssParameters.GetInstance(signatureAlgorithmID.Parameters);
                    AlgorithmIdentifier mgfDigAlg = AlgorithmIdentifier.GetInstance(
                        sigParams.MaskGenAlgorithm.Parameters);

                    FipsRsa.PssSignatureParameters pssParams = FipsRsa.Pss
                        .WithDigest((FipsDigestAlgorithm)Utils.digestTable[sigParams.HashAlgorithm.Algorithm])
                        .WithMgfDigest((FipsDigestAlgorithm)Utils.digestTable[mgfDigAlg.Algorithm])
                        .WithSaltLength(sigParams.SaltLength.IntValueExact);

                    return CreateVerifierFactory(signatureAlgorithmID, verifierService.CreateVerifierFactory(pssParams),
                        m_certificate);
                }
                else if (PkixVerifierFactoryProvider.pkcs1Table.Contains(sigAlgOid))
                {
                    FipsRsa.SignatureParameters rsaParams = FipsRsa.Pkcs1v15.WithDigest(
                        (FipsDigestAlgorithm)PkixVerifierFactoryProvider.pkcs1Table[sigAlgOid]);

                    return CreateVerifierFactory(signatureAlgorithmID, verifierService.CreateVerifierFactory(rsaParams),
                        m_certificate);
                }
                else if (PkcsObjectIdentifiers.RsaEncryption.Equals(sigAlgOid))
                {
                    FipsRsa.SignatureParameters rsaParams = FipsRsa.Pkcs1v15.WithDigest(
                        (FipsDigestAlgorithm)Utils.digestTable[digestAlgorithmID.Algorithm]);

                    return CreateVerifierFactory(signatureAlgorithmID, verifierService.CreateVerifierFactory(rsaParams),
                        m_certificate);
                }
            }
            else if (m_publicKey is AsymmetricDsaPublicKey dsaKey)
            {
                IVerifierFactoryService verifierService = CryptoServicesRegistrar.CreateService(dsaKey);

                FipsDsa.SignatureParameters sigParams;
                if (X9ObjectIdentifiers.IdDsa.Equals(sigAlgOid))
                {
                    sigParams = FipsDsa.Dsa.WithDigest(
                        (FipsDigestAlgorithm)Utils.digestTable[digestAlgorithmID.Algorithm]);
                }
                else
                {
                    sigParams = (FipsDsa.SignatureParameters)PkixVerifierFactoryProvider.dsaTable[sigAlgOid];
                }

                if (sigParams != null)
                {
                    return CreateVerifierFactory(signatureAlgorithmID, verifierService.CreateVerifierFactory(sigParams),
                        m_certificate);
                }
            }
            else if (m_publicKey is AsymmetricECPublicKey ecdsaKey)
            {
                IVerifierFactoryService verifierService = CryptoServicesRegistrar.CreateService(ecdsaKey);

                FipsEC.SignatureParameters sigParams;
                if (X9ObjectIdentifiers.IdECPublicKey.Equals(sigAlgOid))
                {
                    sigParams = FipsEC.Dsa.WithDigest(
                        (FipsDigestAlgorithm)Utils.digestTable[digestAlgorithmID.Algorithm]);
                }
                else
                {
                    sigParams = (FipsEC.SignatureParameters)PkixVerifierFactoryProvider.ecdsaTable[sigAlgOid];
                }

                if (sigParams != null)
                {
                    return CreateVerifierFactory(signatureAlgorithmID, verifierService.CreateVerifierFactory(sigParams),
                        m_certificate);
                }
            }

            throw new ArgumentException("cannot match signature algorithm: " + sigAlgOid);
        }

        private IVerifierFactory<AlgorithmIdentifier> CreateVerifierFactory(AlgorithmIdentifier algorithm,
            IVerifierFactory<IParameters<Algorithm>> baseFactory, X509Certificate certificate)
        {
            if (certificate == null)
                return new PkixVerifierFactory(algorithm, baseFactory);

            return new PkixDatedVerifierFactory(algorithm, baseFactory, certificate);
        }
    }
}
