﻿using System;
using System.Collections;

using Org.BouncyCastle.Asn1.BC;
using Org.BouncyCastle.Asn1.Nist;
using Org.BouncyCastle.Asn1.Pkcs;
using Org.BouncyCastle.Asn1.X509;
using Org.BouncyCastle.Asn1.X9;
using Org.BouncyCastle.Cert;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Asymmetric;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.General;
using Org.BouncyCastle.Utilities;

namespace Org.BouncyCastle.Operators
{
    /// <summary>
    /// Provider class which supports dynamic creation of signature verifiers.
    /// </summary>
	public class PkixVerifierFactoryProvider
        : IVerifierFactoryProvider<AlgorithmIdentifier>
    {
        internal static readonly IDictionary pkcs1Table = Platform.CreateHashtable();
        internal static readonly IDictionary ecdsaTable = Platform.CreateHashtable();
        internal static readonly IDictionary dsaTable = Platform.CreateHashtable();

        static PkixVerifierFactoryProvider()
        {
            pkcs1Table.Add(PkcsObjectIdentifiers.Sha1WithRsaEncryption, FipsShs.Sha1);
            pkcs1Table.Add(PkcsObjectIdentifiers.Sha224WithRsaEncryption, FipsShs.Sha224);
            pkcs1Table.Add(PkcsObjectIdentifiers.Sha256WithRsaEncryption, FipsShs.Sha256);
            pkcs1Table.Add(PkcsObjectIdentifiers.Sha384WithRsaEncryption, FipsShs.Sha384);
            pkcs1Table.Add(PkcsObjectIdentifiers.Sha512WithRsaEncryption, FipsShs.Sha512);

            ecdsaTable.Add(X9ObjectIdentifiers.ECDsaWithSha1, FipsEC.Dsa.WithDigest(FipsShs.Sha1));
            ecdsaTable.Add(X9ObjectIdentifiers.ECDsaWithSha224, FipsEC.Dsa.WithDigest(FipsShs.Sha224));
            ecdsaTable.Add(X9ObjectIdentifiers.ECDsaWithSha256, FipsEC.Dsa.WithDigest(FipsShs.Sha256));
            ecdsaTable.Add(X9ObjectIdentifiers.ECDsaWithSha384, FipsEC.Dsa.WithDigest(FipsShs.Sha384));
            ecdsaTable.Add(X9ObjectIdentifiers.ECDsaWithSha512, FipsEC.Dsa.WithDigest(FipsShs.Sha512));

            dsaTable.Add(X9ObjectIdentifiers.IdDsaWithSha1, FipsDsa.Dsa.WithDigest(FipsShs.Sha1));
            dsaTable.Add(NistObjectIdentifiers.DsaWithSha224, FipsDsa.Dsa.WithDigest(FipsShs.Sha224));
            dsaTable.Add(NistObjectIdentifiers.DsaWithSha256, FipsDsa.Dsa.WithDigest(FipsShs.Sha256));
            dsaTable.Add(NistObjectIdentifiers.DsaWithSha384, FipsDsa.Dsa.WithDigest(FipsShs.Sha384));
            dsaTable.Add(NistObjectIdentifiers.DsaWithSha512, FipsDsa.Dsa.WithDigest(FipsShs.Sha512));
        }

        private readonly IAsymmetricPublicKey m_publicKey;
        private readonly X509Certificate m_certificate;

        /// <summary>
        /// Base constructor - specify the public key to be used in verification.
        /// </summary>
        /// <param name="publicKey">The public key to be used in creating verifiers provided by this object.</param>
		public PkixVerifierFactoryProvider(IAsymmetricPublicKey publicKey)
        {
            m_publicKey = publicKey;
            m_certificate = null;
        }

        /// <summary>
        /// Base constructor - specify the certificate containing the public key to be used in verification.
        /// </summary>
        /// <param name="certificate">The certificate holding the public key to be used.</param>
        public PkixVerifierFactoryProvider(X509Certificate certificate)
        {
            m_publicKey = certificate.GetPublicKey();
            m_certificate = certificate;
        }

        /// <summary>
        /// Return a verifier factory that produces verifiers conforming to algorithmDetails.
        /// </summary>
        /// <param name="algorithmDetails">The configuration parameters for verifiers produced by the resulting factory.</param>
        /// <returns>A new verifier factory.</returns>
        public IVerifierFactory<AlgorithmIdentifier> CreateVerifierFactory(AlgorithmIdentifier algorithmDetails)
        {
            var sigAlgOid = algorithmDetails.Algorithm;

            if (m_publicKey is AsymmetricRsaPublicKey rsaKey)
            {
                IVerifierFactoryService verifierService = CryptoServicesRegistrar.CreateService(rsaKey);

                if (PkcsObjectIdentifiers.IdRsassaPss.Equals(sigAlgOid))
                {
                    RsassaPssParameters sigParams = RsassaPssParameters.GetInstance(algorithmDetails.Parameters);
                    AlgorithmIdentifier mgfDigAlg = AlgorithmIdentifier.GetInstance(
                        sigParams.MaskGenAlgorithm.Parameters);

                    FipsRsa.PssSignatureParameters pssParams = FipsRsa.Pss
                        .WithDigest((FipsDigestAlgorithm)Utils.digestTable[sigParams.HashAlgorithm.Algorithm])
                        .WithMgfDigest((FipsDigestAlgorithm)Utils.digestTable[mgfDigAlg.Algorithm])
                        .WithSaltLength(sigParams.SaltLength.IntValueExact);

                    return CreateVerifierFactory(algorithmDetails, verifierService.CreateVerifierFactory(pssParams),
                        m_certificate);
                }
                else if (pkcs1Table.Contains(sigAlgOid))
                {
                    FipsRsa.SignatureParameters rsaParams = FipsRsa.Pkcs1v15.WithDigest(
                        (FipsDigestAlgorithm)pkcs1Table[sigAlgOid]);

                    return CreateVerifierFactory(algorithmDetails, verifierService.CreateVerifierFactory(rsaParams),
                        m_certificate);
                }
            }
            else if (m_publicKey is AsymmetricDsaPublicKey dsaKey)
            {
                IVerifierFactoryService verifierService = CryptoServicesRegistrar.CreateService(dsaKey);

                FipsDsa.SignatureParameters sigParams = (FipsDsa.SignatureParameters)dsaTable[sigAlgOid];

                return CreateVerifierFactory(algorithmDetails, verifierService.CreateVerifierFactory(sigParams),
                    m_certificate);
            }
            else if (m_publicKey is AsymmetricECPublicKey ecdsaKey)
            {
                IVerifierFactoryService verifierService = CryptoServicesRegistrar.CreateService(ecdsaKey);

                FipsEC.SignatureParameters sigParams = (FipsEC.SignatureParameters)ecdsaTable[sigAlgOid];

                return CreateVerifierFactory(algorithmDetails, verifierService.CreateVerifierFactory(sigParams),
                    m_certificate);
            }
            else if (m_publicKey is AsymmetricSphincsPublicKey sphincsKey)
            {
                IVerifierFactoryService verifierService = CryptoServicesRegistrar.CreateService(sphincsKey);

                Sphincs.SignatureParameters sphincsParams = null;
                if (BCObjectIdentifiers.sphincs256_with_SHA512.Equals(sigAlgOid))
                {
                    sphincsParams = Sphincs.Sphincs256;
                }
                else if (BCObjectIdentifiers.sphincs256_with_SHA3_512.Equals(sigAlgOid))
                {
                    sphincsParams = Sphincs.Sphincs256.WithDigest(FipsShs.Sha3_512);
                }

                if (sphincsParams != null)
                {
                    var factory = verifierService.CreateVerifierFactory(sphincsParams);
                    return CreateVerifierFactory(algorithmDetails, factory, m_certificate);
                }
            }

            throw new ArgumentException("cannot match signature algorithm: " + sigAlgOid);
        }

        private IVerifierFactory<AlgorithmIdentifier> CreateVerifierFactory(AlgorithmIdentifier algorithm,
            IVerifierFactory<IParameters<Algorithm>> baseFactory, X509Certificate certificate)
        {
            if (certificate == null)
                return new PkixVerifierFactory(algorithm, baseFactory);

            return new PkixDatedVerifierFactory(algorithm, baseFactory, certificate);
        }
    }
}
